from typing import Callable

from mongorepo.base import Access
from mongorepo.handlers import _handle_async_mongo_repository


def async_mongo_repository(
    cls: type | None = None,
    add: bool = True,
    add_batch: bool = True,
    get: bool = True,
    get_list: bool = True,
    get_all: bool = True,
    update: bool = True,
    delete: bool = True,
    integer_fields: list[str] | None = None,
    array_fields: list[str] | None = None,
    method_access: Access | None = None,
) -> type | Callable:
    """
    ## Async MongoDB repository decorator

    * decorated class must provide `Meta` class inside
    with variables "dto"(dataclass interface) and
    `collection` (motor collection: `AsyncIOMotorCollection`).

    * You can also provide `index` field that can be just a string name of the field
    which you want to make index field or it can be instance of `mongorepo.Index`
    with extended settings

    * You can use `method_access` to make all methods
    private, protected or public (use `mongorepo.Access`)

    * Add `array_fields` to params to extend repository with methods
    that related to list fields in your dto type

    {field}__pop, {field}__list, {field}__remove, {field__append}

    * Add `integer_fields` to params to extend repository with methods
    that related to integer fields in your dto type

    increment_{field}, decrement_{field}

    ## Example

    ```
    @async_mongo_repository(method_access=mongorepo.Access.PROTECTED)
    class MongoRepository:
        class Meta:
            dto = ExampleDTO
            collection: AsyncIOMotorCollection = db["example"]
            index = mongorepo.Index(field="id_field")

    r = MongoRepository()

    await r.add(ExampleDTO(key="value"))

    example = await repo.get(key="value")  # ExampleDTO(key="value")

    ```

    """

    def wrapper(cls) -> type:
        return _handle_async_mongo_repository(
            cls=cls,
            add=add,
            update=update,
            get_all=get_all,
            get_list=get_list,
            get=get,
            delete=delete,
            add_batch=add_batch,
            integer_fields=integer_fields,
            array_fields=array_fields,
            method_access=method_access,
        )

    if cls is None:
        return wrapper

    return wrapper(cls)



from typing import Callable

from mongorepo._methods.arrays import _pop_list_method, _update_list_field_method
from mongorepo._methods.crud import (
    _add_method,
    _delete_method,
    _get_all_method,
    _get_list_method,
    _get_method,
    _add_batch_method,
    _update_method,
)
from mongorepo._methods.integer import _update_integer_field_method


METHOD_NAME__CALLABLE: dict[str, Callable] = {
    'get': _get_method,
    'add': _add_method,
    'update': _update_method,
    'delete': _delete_method,
    'get_list': _get_list_method,
    'get_all': _get_all_method,
    'pop': _pop_list_method,
    'add_batch': _add_batch_method,
    'update_list': _update_list_field_method,
    'update_integer': _update_integer_field_method,
}

__all__ = ['METHOD_NAME__CALLABLE']



from typing import Callable
from mongorepo.asyncio._methods.arrays import (
    _pop_list_method_async,
    _update_list_field_method_async,
)
from mongorepo.asyncio._methods.crud import (
    _add_method_async,
    _delete_method_async,
    _get_all_method_async,
    _get_list_method_async,
    _get_method_async,
    _add_batch_method_async,
    _update_method_async,
)
from mongorepo.asyncio._methods.integer import _update_integer_field_method_async


METHOD_NAME__CALLABLE: dict[str, Callable] = {
    'get': _get_method_async,
    'add': _add_method_async,
    'update': _update_method_async,
    'delete': _delete_method_async,
    'get_list': _get_list_method_async,
    'get_all': _get_all_method_async,
    'add_batch': _add_batch_method_async,
    'update_list': _update_list_field_method_async,
    'pop': _pop_list_method_async,
    'update_integer': _update_integer_field_method_async,
}

__all__ = ['METHOD_NAME__CALLABLE']



import inspect
from dataclasses import is_dataclass
from types import UnionType
from typing import Any, Callable, NoReturn, Optional, TypeVar, get_args, get_origin

import pymongo
from pymongo.collection import Collection

from mongorepo import DTO, Access, Index
from mongorepo import exceptions


def _get_collection_and_dto(cls: type, raise_exceptions: bool = True) -> dict[str, Any]:
    """Collect `dto` and `collection` attributes from `Meta` class"""
    attributes: dict[str, Any] = {}
    meta = get_meta(cls)

    dto = meta.__dict__.get('dto', None)
    if not dto and raise_exceptions:
        raise exceptions.NoDTOTypeException
    attributes['dto'] = dto

    collection = meta.__dict__.get('collection', None)
    if collection is None and raise_exceptions:
        raise exceptions.NoCollectionException
    attributes['collection'] = collection

    return attributes


def _get_meta_attributes(cls, raise_exceptions: bool = True) -> dict[str, Any]:
    """
    Collect all available attributes from `Meta` class
    """
    attributes: dict[str, Any] = _get_collection_and_dto(
        cls=cls, raise_exceptions=raise_exceptions,
    )
    meta = get_meta(cls)

    index: Index | str | None = getattr(meta, 'index', None)
    attributes['index'] = index

    method_access: Access | None = getattr(meta, 'method_access', None)
    attributes['method_access'] = method_access

    substitute: dict[str, str] | None = getattr(meta, 'substitute', None)
    attributes['substitute'] = substitute

    id_field: str | None = getattr(meta, 'id_field', None)
    attributes['id_field'] = id_field

    return attributes


def get_meta(cls: type) -> Any:
    """
    Tries to get `Meta` class for the class or raises an exception
    """
    meta = cls.__dict__.get('Meta', None)
    if not meta:
        raise exceptions.NoMetaException
    if not isinstance(meta, type):
        raise exceptions.NoMetaException
    return meta


def _get_validated_type_hint(hint: Any, get_type: bool = False) -> Any:
    args = get_args(hint)
    if len(args) > 2:
        raise exceptions.TypeHintException(message=f'Too many arguments for type hint: {hint}')
    if len(args) == 2 and get_origin(hint) is not dict:
        if type(None) not in args:
            raise exceptions.TypeHintException(message=f'"|" allowed only with NoneType {hint}')
    for arg in args:
        if get_origin(arg) is list:
            raise exceptions.TypeHintException(
                message=f'List must have only one argument as type hint {hint}:{arg}'
            )
        if type(arg) is UnionType or type(hint) is Optional:
            raise exceptions.TypeHintException(message=f'Invalid type hint {hint}:{arg}')
        elif is_dataclass(hint):
            get_dto_type_hints(hint)
    if get_type:
        if get_origin(hint) is list:
            return get_origin(hint)
        if args:
            for arg in args:
                if type(arg) is not None:
                    return get_origin(arg)
    return hint


def get_dto_type_hints(dto: type[DTO] | DTO, get_types: bool = True) -> dict[str, Any]:
    """Returns dictionary of fields' type hints for a dataclass or dataclass instance"""
    default_values = {}
    for field_name, value in dto.__annotations__.items():
        hint = _get_validated_type_hint(value, get_type=get_types)
        default_values[field_name] = hint if get_types else value
    return default_values


def create_index(index: Index | str, collection: Collection) -> None:
    """
    ### Creates an index for the collection
    * index parameter can be string or `mongorepo.Index`
    * If index is string, creates standard mongodb index
    * If it's `mongorepo.Index` creates index with user's settings
    """
    if isinstance(index, str):
        collection.create_index(index)
        return
    index_name = f'index_{index.field}'
    if index.name:
        index_name = index.name
    direction = pymongo.DESCENDING if index.desc else pymongo.ASCENDING
    collection.create_index(
        [(index.field, direction)],
        name=index_name,
        unique=index.unique,
    )


def get_prefix(access: Access | None, cls: type | None = None) -> str:
    """
    Returns string prefix according to Access value,
    * it can be `'_'`, `'__'`, `_{cls.__name__}__` or `''`
    """
    match access:
        case Access.PRIVATE:
            prefix = f'_{cls.__name__}__' if cls else '__'
        case Access.PROTECTED:
            prefix = '_'
        case Access.PUBLIC | None:
            prefix = ''
    return prefix


def _get_dto_from_origin(cls: type) -> Any:
    """
    Tries to get `dto` from origin of the class or raises an exception

    ```
    class A[UserDTO]: ...

    _get_dto_from_origin(A)  # UserDTO

    ```
    """
    try:
        if not hasattr(cls, '__orig_bases__'):
            raise exceptions.NoDTOTypeException
        dto: type = get_args(cls.__orig_bases__[0])[0]
    except IndexError:
        raise exceptions.NoDTOTypeException
    if dto is DTO:
        raise exceptions.NoDTOTypeException
    if not is_dataclass(dto):
        raise exceptions.NotDataClass

    return dto


def convert_to_dto(dto_type: type[DTO], dct: dict[str, Any]) -> DTO:
    """
    Converts document to dto, does not include mongodb `_id`
    """
    dct.pop('_id') if dct.get('_id', None) else ...
    return dto_type(**dct)


def convert_to_dto_with_id(
    id_field: str,
) -> Callable:
    """
    Converts document to dto,
    includes mongodb `_id` allows to set specific field where to store `_id`
    """
    def wrapper(dto_type: type[DTO], dct: dict[str, Any]) -> DTO:
        dct[id_field] = str(dct.pop('_id'))
        return dto_type(**dct)
    return wrapper


def recursive_convert_to_dto(dto_type: type[DTO], id_field: str | None = None) -> Callable:
    def decorator(dto_type: type[DTO], dct: dict[str, Any]) -> DTO:
        def wrapper(
            dto_type: type[DTO], dct: dict[str, Any], to_dto: bool = False,
        ) -> dict[str, Any] | DTO:
            type_hints = get_dto_type_hints(dto_type, get_types=False)
            data = {}
            for key, value in dct.items():
                if is_dataclass(type_hints.get(key, None)):
                    data[key] = wrapper(type_hints.get(key, None), value, to_dto=True)
                elif get_origin(type_hints.get(key, None)) is list and isinstance(value, list):
                    args = get_args(type_hints.get(key, None))
                    if is_dataclass(args[0]):
                        data[key] = [
                            wrapper(args[0], v, to_dto=True) for v in value  # type: ignore
                        ]
                    else:
                        data[key] = value
                else:
                    data[key] = value
            return dto_type(**data) if to_dto else data
        data = {}
        if id_field is not None:
            data[id_field] = str(dct.pop('_id'))
        else:
            dct.pop('_id') if dct.get('_id', None) else ...
        data.update(wrapper(dto_type, dct))  # type: ignore
        return dto_type(**data)
    return decorator


def _get_converter(dto_type: type[DTO], id_field: str | None = None) -> Callable:
    """
    Returns proper converter based on type hints of the dto
    """
    converter = convert_to_dto
    r = _has_dataclass_fields(dto_type=dto_type)
    if r:
        converter = recursive_convert_to_dto(dto_type, id_field)
    elif id_field is not None:
        converter = convert_to_dto_with_id(id_field=id_field)
    return converter


def _has_dataclass_fields(dto_type: type[DTO]) -> bool:
    type_hints = get_dto_type_hints(dto_type, get_types=False)
    for value in type_hints.values():
        if is_dataclass(value):
            return True
        elif get_origin(value) is list:
            args = get_args(value)
            if is_dataclass(args[0]):
                return True
    return False


def get_dataclass_fields(
    dto_type: type[DTO],
    only_dto_types: bool = False,
) -> dict[str, type[DTO]]:
    """
    Returns dictionary of fields which has dataclasses as type hints or as arguments

    `only_types=True` to get only dto types,
    instead of `{"example": list[ExampleDTO]}` get `{"example": ExampleDTO}`
    """
    dataclass_fields = {}
    type_hints = get_dto_type_hints(dto_type, get_types=False)
    for key, value in type_hints.items():
        if is_dataclass(value):
            dataclass_fields[key] = value
        elif get_origin(value) is list:
            args = get_args(value)
            if is_dataclass(args[0]):
                dataclass_fields[key] = args[0] if only_dto_types else value
    return dataclass_fields


def raise_exc(exc: Exception) -> NoReturn:
    """Allows to write one-lined exceptions"""
    raise exc


def _validate_method_annotations(method: Callable) -> None:
    if not method.__annotations__:
        raise exceptions.MongoRepoException(message=f'No type hints for {method.__name__}()')
    if 'return' not in method.__annotations__:
        raise exceptions.MongoRepoException(
            message=f'return type is not specified for "{method}" method',
        )
    params = inspect.signature(method).parameters
    if list(params)[0] != 'self':
        raise exceptions.MongoRepoException(
            message=f'First parameter must be self: "{method}" method',
        )
    for param, type_hint in params.items():
        if param == 'self':
            continue
        if type_hint == inspect._empty and type_hint.kind not in [
            inspect.Parameter.VAR_KEYWORD, inspect.Parameter.VAR_POSITIONAL
        ]:
            raise exceptions.MongoRepoException(
                message=f'Parameter "{param}" does not have type hint',
            )


def replace_typevars(func: Callable, typevar: Any) -> None:
    for param, anno in func.__annotations__.items():
        if isinstance(anno, TypeVar):
            func.__annotations__[param] = typevar



from dataclasses import Field, dataclass
from enum import Enum
from typing import Any, ClassVar, Protocol, TypeVar


class DataclassInstance(Protocol):
    __dataclass_fields__: ClassVar[dict[str, Field[Any]]]


DTO = TypeVar('DTO', bound=DataclassInstance)
_DTOField = TypeVar('_DTOField', bound=DataclassInstance)


@dataclass(repr=False, slots=True, eq=False)
class Index:
    """Allows to specify extra information about mongodb index"""
    field: str
    name: str | None = None
    desc: bool = True
    unique: bool = False


class Access(int, Enum):
    """
    Use to indicate method access in a repository, see also `mongorepo.utils.get_prefix` function
    """
    PUBLIC = 0
    PROTECTED = 1
    PRIVATE = 2



from typing import Callable

from motor.motor_asyncio import AsyncIOMotorCollection

from mongorepo.utils import raise_exc
from mongorepo import DTO, exceptions


def _update_integer_field_method_async(
    dto_type: type[DTO],
    collection: AsyncIOMotorCollection,
    field_name: str,
    _weight: int = 1,
) -> Callable:
    async def update_integer_field(self, weight: int | None = None, **filters) -> None:
        w = weight if weight is not None else _weight
        document = await collection.find_one_and_update(
            filter=filters, update={'$inc': {field_name: w}}, return_document=True,
        )
        raise_exc(exceptions.NotFoundException(**filters)) if not document else ...
    return update_integer_field



from dataclasses import asdict
from typing import Any, Callable, Iterable

from bson import ObjectId
from pymongo.collection import Collection
from mongorepo.utils import _get_converter
from mongorepo import DTO


def _add_method(
    dto_type: type[DTO],
    collection: Collection,
    id_field: str | None = None,
) -> Callable:

    def add(self, dto: DTO) -> DTO:
        collection.insert_one(asdict(dto))
        return dto

    def add_and_return_with_id(self, dto: DTO) -> DTO:
        object_id = ObjectId()
        dto.__dict__[id_field] = str(object_id)  # type: ignore
        collection.insert_one({**asdict(dto), '_id': object_id})
        return dto

    if not id_field:
        return add
    return add_and_return_with_id


def _add_batch_method(
    dto_type: type[DTO],
    collection: Collection,
    id_field: str | None = None,
) -> Callable:

    def add(self, dto_list: list[DTO]) -> None:
        collection.insert_many(asdict(d) for d in dto_list)

    def add_and_return_with_id(self, dto_list: list[DTO]) -> None:
        batch: list[dict[str, Any]] = []
        for dto in dto_list:
            object_id = ObjectId()
            dto.__dict__[id_field] = str(object_id)  # type: ignore
            batch.append({**asdict(dto), '_id': object_id})
        collection.insert_many(batch)

    if id_field is not None:
        return add
    return add_and_return_with_id


def _get_list_method(
    dto_type: type[DTO],
    collection: Collection,
    id_field: str | None = None,
):
    to_dto = _get_converter(dto_type, id_field=id_field)

    def get_list(self, offset: int = 0, limit: int = 20) -> list[DTO]:
        cursor = collection.find().skip(offset).limit(limit)
        return [to_dto(dto_type, doc) for doc in cursor]
    return get_list


def _get_all_method(
    dto_type: type[DTO],
    collection: Collection,
    id_field: str | None = None
) -> Callable:
    to_dto = _get_converter(dto_type=dto_type, id_field=id_field)

    def get_all(self, **filters: Any) -> Iterable[DTO]:
        cursor = collection.find(filters)
        for dct in cursor:
            yield to_dto(dto_type, dct)
    return get_all


def _update_method(
    dto_type: type[DTO],
    collection: Collection,
    id_field: str | None = None,
) -> Callable:
    to_dto = _get_converter(dto_type, id_field=id_field)

    def update(self, dto: DTO, **filters: Any) -> DTO | None:
        data: dict[str, dict[str, Any]] = {'$set': {}}
        for field, value in asdict(dto).items():
            data['$set'][field] = value
        updated_document: dict[str, Any] | None = collection.find_one_and_update(
            filter=filters, update=data, return_document=True,
        )
        return to_dto(dto_type, updated_document) if updated_document else None
    return update


def _delete_method(dto_type: type[DTO], collection: Collection) -> Callable:
    def delete(self, **filters: Any) -> bool:
        deleted = collection.find_one_and_delete(filters)
        return True if deleted else False
    return delete


def _get_method(
    dto_type: type[DTO],
    collection: Collection,
    id_field: str | None = None,
) -> Callable:
    to_dto = _get_converter(dto_type, id_field=id_field)

    def get(self, **filters: Any) -> DTO | None:
        result = collection.find_one(filters)
        return to_dto(dto_type, result) if result else None
    return get



import inspect
from typing import Callable

from mongorepo.asyncio.utils import _run_asyncio_create_index
from mongorepo.base import Access
from mongorepo.setters import (
    _set_array_fields_methods,
    _set_crud_methods,
    _set_integer_fields_methods,
)
from mongorepo.utils import _get_meta_attributes, raise_exc, create_index
from mongorepo import exceptions
from mongorepo._methods import METHOD_NAME__CALLABLE
from mongorepo._substitute import _substitute_method
from mongorepo.asyncio._methods import METHOD_NAME__CALLABLE as METHOD_NAME__CALLABLE_ASYNC


def _handle_mongo_repository(
    cls,
    add: bool,
    get: bool,
    add_batch: bool,
    get_all: bool,
    update: bool,
    delete: bool,
    get_list: bool,
    integer_fields: list[str] | None,
    array_fields: list[str] | None,
    method_access: Access | None,
) -> type:
    """
    Calls for functions that set different methods and attributes to the class
    """
    attributes = _get_meta_attributes(cls, raise_exceptions=False)
    collection = attributes['collection']
    index = attributes['index']

    _set_crud_methods(
        cls,
        add=add,
        get=get,
        get_all=get_all,
        get_list=get_list,
        update=update,
        delete=delete,
        add_batch=add_batch,
        method_access=method_access,
    )

    if index is not None:
        create_index(index=index, collection=collection)

    if integer_fields is not None:
        _set_integer_fields_methods(cls, integer_fields=integer_fields, method_access=method_access)

    if array_fields is not None:
        _set_array_fields_methods(cls, array_fields=array_fields, method_access=method_access)

    return cls


def _handle_async_mongo_repository(
    cls,
    add: bool,
    add_batch: bool,
    get: bool,
    get_all: bool,
    get_list: bool,
    update: bool,
    delete: bool,
    integer_fields: list[str] | None,
    array_fields: list[str] | None,
    method_access: Access | None,
) -> type:
    """
    Calls for functions that set different async methods and attributes to the class
    """
    attributes = _get_meta_attributes(cls, raise_exceptions=False)
    collection = attributes['collection']
    index = attributes['index']

    _set_crud_methods(
        cls,
        add=add,
        get=get,
        get_all=get_all,
        update=update,
        delete=delete,
        add_batch=add_batch,
        method_access=method_access,
        get_list=get_list,
        async_methods=True,
    )

    if integer_fields is not None:
        _set_integer_fields_methods(
            cls, integer_fields=integer_fields, async_methods=True, method_access=method_access,
        )

    if array_fields is not None:
        _set_array_fields_methods(
            cls, array_fields=array_fields, async_methods=True, method_access=method_access,
        )

    if index is not None:
        _run_asyncio_create_index(index=index, collection=collection)
    return cls


def _handle_implements(
    base_cls: type,
    cls: type,
    **substitute: str | Callable,
) -> type:
    attrs = _get_meta_attributes(cls)
    if not substitute:
        substitute = attrs['substitute'] if attrs['substitute'] is not None else raise_exc(
            exceptions.MongoRepoException(message='No "substitute" in Meta class')
        )
    dto_type = attrs['dto']
    collection = attrs['collection']
    id_field = attrs['id_field']
    for mongorepo_method_name, _generic_method in substitute.items():
        if inspect.isfunction(_generic_method) or inspect.ismethod(_generic_method):
            generic_method: Callable = _generic_method  # type: ignore
        elif isinstance(_generic_method, str):
            generic_method: Callable | None = getattr(
                base_cls, _generic_method, None
            )
            if generic_method is None or not inspect.isfunction(generic_method):
                raise exceptions.InvalidMethodNameException(_generic_method)

        is_async: bool = inspect.iscoroutinefunction(generic_method)

        if mongorepo_method_name not in METHOD_NAME__CALLABLE:
            raise exceptions.InvalidMethodNameException(mongorepo_method_name)
        if is_async:
            mongorepo_method: Callable = METHOD_NAME__CALLABLE_ASYNC[mongorepo_method_name]
        else:
            mongorepo_method: Callable = METHOD_NAME__CALLABLE[
                mongorepo_method_name
            ]

        setattr(
            cls,
            generic_method.__name__,
            _substitute_method(
                mongorepo_method,
                generic_method,
                dto_type,
                collection,
                id_field=id_field,
            ),
        )

    return cls



import inspect
from dataclasses import is_dataclass
from inspect import Parameter
from typing import Any, Callable, TypeVar

from motor.motor_asyncio import AsyncIOMotorCollection
from pymongo.collection import Collection

from mongorepo.base import DTO
from mongorepo.utils import _validate_method_annotations, replace_typevars


def _substitute_method(
    mongorepo_method: Callable,
    generic_method: Callable,
    dto: type[DTO],
    collection: Collection | AsyncIOMotorCollection,
    id_field: str | None = None,
) -> Callable:
    if id_field in mongorepo_method.__annotations__:
        mongorepo_method = mongorepo_method(dto_type=dto, collection=collection, id_field=id_field)
    else:
        mongorepo_method = mongorepo_method(dto_type=dto, collection=collection)

    is_async = inspect.iscoroutinefunction(generic_method)
    _validate_method_annotations(generic_method)

    def func(self, *args, **kwargs) -> Any:
        required_params = _manage_params(
            mongorepo_method, generic_method, *args, **kwargs,
        )
        return mongorepo_method(self, **required_params)

    async def async_func(self, *args, **kwargs) -> Any:
        required_params = _manage_params(
            mongorepo_method, generic_method, *args, **kwargs,
        )
        return await mongorepo_method(self, **required_params)

    new_method = async_func if is_async else func

    new_method.__annotations__ = generic_method.__annotations__
    new_method.__name__ = generic_method.__name__
    new_method.__annotations__['return'] = dto | None

    replace_typevars(new_method, dto)

    return new_method


def _manage_params(
    mongorepo_method: Callable,
    generic_method: Callable,
    *args,
    **kwargs,
) -> dict[str, Any]:
    """Return required params for `mongorepo` method, do not pass `self` in *args or **kwargs"""
    _kwargs = {}
    gen_params = dict(inspect.signature(generic_method).parameters)
    gen_params.pop('self')
    mongo_params = dict(inspect.signature(mongorepo_method).parameters)
    mongo_params.pop('self')

    i = 0
    for param in gen_params.values():
        try:
            arg = args[i]
            i += 1
            _kwargs[param.name] = arg
        except IndexError:
            break

    for param in gen_params.values():
        if param.name not in _kwargs and param.name not in kwargs:
            raise TypeError(
                f'{generic_method.__name__}() '
                f'missing required keyword argument: \'{param.name}\''
            )
        if param.name in kwargs and param.name in _kwargs:
            raise TypeError(
                f'{generic_method.__name__}() keyword arguments repeated: {param.name}'
            )
        if param.name in _kwargs:
            continue
        _kwargs[param.name] = kwargs[param.name]

    result = {}
    extra: bool = False
    for mongo_param in mongo_params.values():
        if mongo_param.kind is Parameter.VAR_KEYWORD:
            extra = True
            break
        if isinstance(mongo_param.annotation, TypeVar) or is_dataclass(mongo_param.annotation):
            for key, value in _kwargs.items():
                if is_dataclass(value):
                    result[mongo_param.name] = value
                    _kwargs[key] = None
                    break
        else:
            for key, value in _kwargs.items():
                if isinstance(value, mongo_param.annotation):
                    result[mongo_param.name] = value
                    _kwargs[key] = None
                    break
    if extra:
        for key, value in _kwargs.items():
            if value is not None:
                result[key] = value
    return result



from dataclasses import asdict, is_dataclass
from typing import Any, Callable

from motor.motor_asyncio import AsyncIOMotorCollection

from mongorepo.utils import _get_converter, get_dataclass_fields, raise_exc
from mongorepo import DTO, exceptions
from mongorepo.base import _DTOField


def _get_list_of_field_values_method_async(
    dto_type: type[DTO], collection: AsyncIOMotorCollection, field_name: str,
) -> Callable:
    dataclass_fields = get_dataclass_fields(dto_type=dto_type, only_dto_types=True)
    field_type = dataclass_fields.get(field_name, None)

    async def get_list_dto(
        self, offset: int, limit: int, **filters,
    ) -> list[_DTOField]:  # type: ignore
        document = await collection.find_one(
            filters, {field_name: {'$slice': [offset, limit]}},
        )
        raise_exc(exceptions.NotFoundException(**filters)) if not document else ...
        return [to_dto(field_type, d) for d in document[field_name]]

    async def get_list(self, offset: int, limit: int, **filters) -> list[Any]:
        document = await collection.find_one(
            filters, {field_name: {'$slice': [offset, limit]}},
        )
        raise_exc(exceptions.NotFoundException(**filters)) if not document else ...
        return document[field_name]

    if is_dataclass(field_type):
        to_dto = _get_converter(field_type)
        return get_list_dto
    return get_list


def _update_list_field_method_async(
    dto_type: type[DTO],
    collection: AsyncIOMotorCollection,
    field_name: str,
    command: str = '$push',
) -> Callable:
    dataclass_fields = get_dataclass_fields(dto_type=dto_type, only_dto_types=True)
    field_type = dataclass_fields.get(field_name, None)

    async def update_list(self, value: Any, **filters) -> None:
        value = value if not is_dataclass(field_type) else asdict(value)
        document = await collection.update_one(
            filter=filters, update={command: {field_name: value}},
        )
        raise_exc(exceptions.NotFoundException(**filters)) if not document else ...
    return update_list


def _pop_list_method_async(
    dto_type: type[DTO],
    collection: AsyncIOMotorCollection,
    field_name: str,
) -> Callable:
    dataclass_fields = get_dataclass_fields(dto_type=dto_type, only_dto_types=True)
    field_type = dataclass_fields.get(field_name, None)

    async def pop_list(self, **filters) -> Any | None:
        document = await collection.find_one_and_update(
            filter=filters, update={'$pop': {field_name: 1}},
        )
        raise_exc(exceptions.NotFoundException(**filters)) if not document else ...
        return document[field_name][-1] if document else None

    async def pop_list_dto(self, **filters) -> Any | None:
        document = await collection.find_one_and_update(
            filter=filters, update={'$pop': {field_name: 1}},
        )
        raise_exc(exceptions.NotFoundException(**filters)) if not document else ...
        return to_dto(field_type, document[field_name][-1]) if document else None

    if is_dataclass(field_type):
        to_dto = _get_converter(dataclass_fields[field_name])
        return pop_list_dto
    return pop_list
